# This file is part of xrayutilities.
#
# xrayutilities is free software; you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation; either version 2 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program; if not, see <http://www.gnu.org/licenses/>.
#
# Copyright (C) 2013-2014 Dominik Kriegner <dominik.kriegner@gmail.com>


"""
class for reading data + header information from tty08 data files

tty08 is a system used at beamline P08 at Hasylab Hamburg and creates simple
ASCII files to save the data. Information is easily read from the multicolumn
data file. the functions below enable also to parse the information of the
header
"""

import glob
import os.path
import re

import numpy
import numpy.lib.recfunctions

from ..exception import InputError
# relative imports from xrayutilities
from .helper import xu_open

re_columns = re.compile(r"/\*H")
re_command = re.compile(r"^/\*C command")
re_comment = re.compile(r"^/\*")
re_date = re.compile(r"^/\*D date")
re_epoch = re.compile(r"^/\*T epoch")
re_initmopo = re.compile(r"^/\*M")


class tty08File(object):

    """
    Represents a tty08 data file. The file is read during the
    Constructor call. This class should work for data stored at
    beamline P08 using the tty08 acquisition system.

    Parameters
    ----------
    filename :  str
        tty08-filename
    mcadir :    str, optional
        directory name of MCA files
    """

    def __init__(self, filename, path=None, mcadir=None):
        self.filename = filename
        if path is None:
            self.full_filename = self.filename
        else:
            self.full_filename = os.path.join(path, self.filename)

        self.Read()

        if mcadir is not None:
            self.mca_directory = mcadir
            self.mca_files = sorted(glob.glob(
                os.path.join(self.mca_directory, '*')))

            if self.mca_files:
                self.ReadMCA()

    def ReadMCA(self):
        self.mca = numpy.empty((len(self.mca_files),
                                numpy.loadtxt(self.mca_files[0]).shape[0]),
                               dtype=numpy.float)
        for i in range(len(self.mca_files)):
            mcadata = numpy.loadtxt(self.mca_files[i])

            self.mca[i, :] = mcadata[:, 1]

            if i == 0:
                if len(mcadata.shape) == 2:
                    self.mca_channels = mcadata[:, 0]
                else:
                    self.mca_channels = numpy.arange(0, mcadata.shape[0])

        mcatemp = self.mca.view([('MCA',
                                  (self.mca.dtype, self.mca.shape[1]))])
        self.data = numpy.lib.recfunctions.merge_arrays([self.data, mcatemp],
                                                        flatten=True)

    def Read(self):
        """
        Read the data from the file
        """

        with xu_open(self.full_filename) as fid:
            # read header
            self.init_mopo = {}
            for line in fid:
                line = line.decode('ascii')

                if re_command.match(line):
                    m = line.split(':')
                    self.scan_command = m[1].strip()
                if re_date.match(line):
                    m = line.split(':', 1)
                    self.scan_date = m[1].strip()
                if re_epoch.match(line):
                    m = line.split(':', 1)
                    self.epoch = float(m[1])
                if re_initmopo.match(line):
                    m = line[3:]
                    m = m.split(';')
                    for e in m:
                        e = e.split('=')
                        self.init_mopo[e[0].strip()] = float(e[1])

                if re_columns.match(line):
                    self.columns = tuple(line.split()[1:])
                    # here all necessary information is read and we can start
                    # reading the data
                    break
            self.data = numpy.loadtxt(fid, comments="/")

        self.data = numpy.rec.fromrecords(self.data, names=self.columns)


def gettty08_scan(scanname, scannumbers, *args, **keyargs):
    """
    function to obtain the angular cooridinates as well as intensity values
    saved in TTY08 datafiles. Especially usefull for reciprocal space map
    measurements, and to combine date from several scans

    further more it is possible to obtain even more positions from
    the data file if more than two string arguments with its names are given

    Parameters
    ----------
    scanname :      str
        name of the scans, for multiple scans this needs to be a template
        string
    scannumbers :   int, tuple or list
        number of the scans of the reciprocal space map

    args :          str, optional
        names of the motors. to read reciprocal space maps measured in coplanar
        diffraction give:

            - `omname`: the name of the omega motor (or its equivalent)
            - `ttname`: the name of the two theta motor (or its equivalent)

    keyargs :       dict, optional
        keyword arguments are passed on to tty08File

    Returns
    -------
    [ang1, ang2, ...] :     list, optional
        angular positions of the center channel of the position sensitive
        detector (numpy.ndarray 1D), omitted if no `args` are given
    MAP :                   ndarray
        All the data values as stored in the data file (includes the
        intensities e.g. MAP['MCA']).

    Examples
    --------
    >>> [om, tt], MAP = xu.io.gettty08_scan('text%05d.dat', 36, 'omega',
    >>>                                     'gamma')
    """

    if isinstance(scannumbers, (list, tuple)):
        scanlist = scannumbers
    else:
        scanlist = list([scannumbers])

    angles = dict.fromkeys(args)
    for key in angles:
        if not isinstance(key, str):
            raise InputError("*arg values need to be strings with motornames")
        angles[key] = numpy.zeros(0)
    buf = numpy.zeros(0)
    MAP = numpy.zeros(0)

    for nr in scanlist:
        scan = tty08File(scanname % nr, **keyargs)
        sdata = scan.data
        if MAP.dtype == numpy.float64:
            MAP.dtype = sdata.dtype
        # append scan data to MAP, where all data are stored
        MAP = numpy.append(MAP, sdata)
        # check type of scan
        for i in range(len(args)):
            motname = args[i]
            scanlength = len(sdata)
            try:
                buf = sdata[motname]
            except ValueError:
                buf = scan.init_mopo[motname] * numpy.ones(scanlength)
            angles[motname] = numpy.concatenate((angles[motname], buf))

    retval = []
    for motname in args:
        # create return values in correct order
        retval.append(angles[motname])

    if not args:
        return MAP
    elif len(args) == 1:
        return retval[0], MAP
    else:
        return retval, MAP



# This file is part of xrayutilities.
#
# xrayutilities is free software; you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation; either version 2 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program; if not, see <http://www.gnu.org/licenses/>.
#
# Copyright (C) 2012,2018 Dominik Kriegner <dominik.kriegner@gmail.com>

# ALSO LOOK AT THE FILE xrayutilities_example_plot_3D_ESRF_ID01.py

import collections.abc
import re

import xrayutilities as xu
from matplotlib.pylab import *

# define root of the local data directory (needed because we assume the data
# path of detector frames from the specfile are not correct anymore)
datadir = '/local/data/path'
repl_n = len(datadir.split('/'))

# define intensity normalizer class to normalize for count time and
# monitor changes: to have comparable absolute intensities set the keyword
# argument av_mon to a fixed value, otherwise different scans can not be
# compared!
xid01_normalizer = xu.IntensityNormalizer(
    'CCD', time='Seconds', mon='exp1', av_mon=10000.0)


def deadpixelkill(ccdraw, f=1.0):
    """
    fill empty "spacer" pixels of the maxipix 2x2 detector
    """
    ccd = ccdraw.astype(float32)
    ccd[255:258, :] = ccd[255, :]/3*f
    ccd[258:261, :] = ccd[260, :]/3*f
    ccd[:, 255:258] = ones((ccd.shape[0], 3)) * (ccd[:, 255])[:, newaxis]/3*f
    ccd[:, 258:261] = ones((ccd.shape[0], 3)) * (ccd[:, 260])[:, newaxis]/3*f
    return ccd


def getmpx4_filetmp(specscan, replace=repl_n, key='ULIMA_mpx4', ext='.edf.gz',
                    datadir=datadir):
    """
    read MaxiPix file template from the scan header.

    Parameters
    ----------
     specscan:  SPECScan object from which the header should be read
     replace:   number of folders which should be replaced by the new data
                directory
     key:       spec-header key name for the file template
     ext:       file extension of the EDF files
     datadir:   new root of the data directory

    Returns
    -------
     file template for the maxipix pictures
    """
    ret = specscan.getheader_element(key)
    ret = xu.utilities.exchange_path(ret, datadir, replace=replace)
    fn = ret[:-len(ext)]
    sfn = fn.split('_')
    return '_'.join(sfn[:-1]) + '_%05d' + ext


def getmono_energy(specscan, key='UMONO', motor='mononrj'):
    """
    read the monochromator energy from the spec-scan header

    Parameters
    ----------
     specscan:  SPECScan object from which the header should be read
     key:       spec-header key name for the monochromator parameters
     motor:     name of the energy motor

    Returns
    -------
     energy, float in eV
    """
    ret = specscan.getheader_element(key)
    motors = ret.split(',')
    re_val_unit = re.compile(r'([0-9.]*)([a-zA-Z]*)')
    for m in motors:
        mn, mv = m.split('=')
        if mn == motor:
            match = re_val_unit.match(mv)
            en = float(match.groups()[0])*1e3
    return en


def rawmap(specfile, scannr, experiment, angdelta=None, U=identity(3),
           norm=True):
    """
    read ccd frames and and convert them in reciprocal space
    angular coordinates are taken from the spec file. A single
    or multiple scan-numbers can be given.
    """

    [mu, eta, phi, nu, delta, ty, tz, ccdn] = xu.io.getspec_scan(
        specfile, scannr, 'mu', 'eta', 'phi', 'nu',
        'del', 'mpxy', 'mpxz', 'mpx4inr')

    idx = 0
    nav = experiment._A2QConversion._area_nav
    roi = experiment._A2QConversion._area_roi

    if not isinstance(scannr, collections.abc.Iterable):
        scannr = [scannr, ]
    for snr in scannr:
        specscan = getattr(specfile, 'scan%d' % snr)
        ccdfiletmp = getmpx4_filetmp(specscan)
        en = getmono_energy(specscan)
        experiment.energy = en

        for j in range(len(specscan.data)):
            i = ccdn[idx]
            # read ccd image from EDF file
            e = xu.io.EDFFile(ccdfiletmp % i)
            ccd = deadpixelkill(e.data)

            # normalize ccd-data (optional)
            # create data for normalization
            if norm:
                d = {'CCD': ccd, 'exp1': specscan.data['exp1'][idx],
                     'Seconds': specscan.data['Seconds'][idx]}
                ccd = xid01_normalizer(d)
            CCD = xu.blockAverage2D(ccd, nav[0], nav[1], roi=roi)

            if idx == 0:
                intensity = zeros((len(ccdn), ) + CCD.shape)

            intensity[idx, :, :] = CCD
            idx += 1

    # transform scan angles to reciprocal space coordinates for all detector
    # pixels
    if angdelta is None:
        qx, qy, qz = experiment.Ang2Q.area(mu, eta, phi, nu, delta,
                                           ty, tz, UB=U)
    else:
        qx, qy, qz = experiment.Ang2Q.area(mu, eta, phi, nu, delta,
                                           ty, tz, delta=angdelta, UB=U)

    return qx, qy, qz, intensity


def gridmap(specfile, scannr, experiment, nx, ny, nz, **kwargs):
    """
    read ccd frames and grid them in reciprocal space
    angular coordinates are taken from the spec file

    **kwargs are passed to the rawmap function
    """

    qx, qy, qz, intensity = rawmap(specfile, scannr, experiment, **kwargs)

    # convert data to rectangular grid in reciprocal space
    gridder = xu.Gridder3D(nx, ny, nz)
    gridder(qx, qy, qz, intensity)

    return gridder.xaxis, gridder.yaxis, gridder.zaxis, gridder.data, gridder



# This file is part of xrayutilities.
#
# xrayutilies is free software; you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation; either version 2 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program; if not, see <http://www.gnu.org/licenses/>.
#
# Copyright (C) 2009 Eugen Wintersberger <eugen.wintersberger@desy.de>
# Copyright (C) 2010-2021 Dominik Kriegner <dominik.kriegner@gmail.com>

import glob
import os.path
import subprocess
import sys
import tempfile
from distutils.command.build_py import build_py
from distutils.command.install import INSTALL_SCHEMES
from distutils.errors import CompileError, DistutilsArgError
from distutils.fancy_getopt import FancyGetopt

import numpy
from setuptools import Extension, find_packages, setup
from setuptools.command.build_ext import build_ext

cliopts = []
cliopts.append(("without-openmp", None, "build without OpenMP support"))

options = FancyGetopt(option_table=cliopts)

# Modify the data install dir to match the source install dir
for scheme in INSTALL_SCHEMES.values():
    scheme['data'] = scheme['purelib']

# first read all the arguments passed to the script
# we need to do this otherwise the --help commands would not work
args = sys.argv[1:]
try:
    # search the arguments for options we would like to use
    # get new args with the custom options stripped away
    args, opts = options.getopt(args)
except DistutilsArgError:
    pass

# get options from command line
without_openmp = False
for opts, values in options.get_option_order():
    if opts == "without-openmp":
        without_openmp = True


def has_flag(compiler, flagname, output_dir=None):
    # see https://bugs.python.org/issue26689
    """Return a boolean indicating whether a flag name is supported on
    the specified compiler.
    """
    with tempfile.NamedTemporaryFile('w', suffix='.c', delete=False) as f:
        f.write('int main (int argc, char **argv) { return 0; }')
        f.close()
        try:
            obj = compiler.compile([f.name], output_dir=output_dir,
                                   extra_postargs=[flagname])
            os.remove(*obj)
        except CompileError:
            return False
        finally:
            os.remove(f.name)
    return True


class build_ext_subclass(build_ext):

    def build_extensions(self):
        c = self.compiler.compiler_type
        # set extra compiler options
        copt = {}

        if c in copt:
            for flag in copt[c]:
                if has_flag(self.compiler, flag, self.build_temp):
                    for e in self.extensions:
                        e.extra_compile_args.append(flag)

        # set openMP compiler options
        if not without_openmp:
            openmpflags = {'msvc': ('/openmp', None),
                           'mingw32': ('-fopenmp', '-fopenmp'),
                           'unix': ('-fopenmp', '-lgomp')}
            if c in openmpflags:
                flag, lib = openmpflags[c]
                if has_flag(self.compiler, flag, self.build_temp):
                    for e in self.extensions:
                        e.extra_compile_args.append(flag)
                        if lib is not None:
                            e.extra_link_args.append(lib)
                        e.define_macros.append(('__OPENMP__', None))

        super().build_extensions()


class build_with_database(build_py):
    def build_database(self):
        dbfilename = os.path.join(self.build_lib, 'xrayutilities',
                                  'materials', 'data', 'elements.db')
        cmd = [sys.executable,
               os.path.join('lib', 'xrayutilities', 'materials',
                            '_create_database.py'),
               dbfilename]
        self.mkpath(os.path.dirname(dbfilename))
        print('building database: {}'.format(dbfilename))
        try:
            if sys.version_info >= (3, 5):
                subprocess.run(cmd, stderr=subprocess.PIPE,
                               stdout=subprocess.PIPE, check=True)
            else:
                subprocess.check_output(cmd)
        except subprocess.CalledProcessError as cpe:
            sys.stdout.buffer.write(cpe.stdout)
            sys.stdout.buffer.write(cpe.stderr)
            raise

    def run(self):
        super().run()
        self.build_database()


cmdclass = {'build_py': build_with_database,
            'build_ext': build_ext_subclass}

with open('README.md') as f:
    long_description = f.read()

extmodul = Extension('xrayutilities.cxrayutilities',
                     sources=glob.glob(os.path.join('src', '*.c')))

with open('lib/xrayutilities/VERSION') as version_file:
    version = version_file.read().strip().replace('\n', '.')

try:
    import sphinx
    from sphinx.setup_command import BuildDoc

    class build_doc(BuildDoc):
        def run(self):
            # make sure the python path is pointing to the newly built
            # code so that the documentation is built on this and not a
            # previously installed version
            build = self.get_finalized_command('build')
            sys.path.insert(0, os.path.abspath(build.build_lib))
            try:
                sphinx.setup_command.BuildDoc.run(self)
            except UnicodeDecodeError:
                print("ERROR: unable to build documentation"
                      " because Sphinx do not handle"
                      " source path with non-ASCII characters. Please"
                      " try to move the source package to another"
                      " location (path with *only* ASCII characters)")
            sys.path.pop(0)

    cmdclass['build_doc'] = build_doc

except ImportError:
    pass

setup(
    name="xrayutilities",
    version=version,
    author="Eugen Wintersberger, Dominik Kriegner",
    description="package for x-ray diffraction data evaluation",
    classifiers=[
        "Programming Language :: C",
        "Programming Language :: Python :: 3.6",
        "Programming Language :: Python :: 3.7",
        "Programming Language :: Python :: 3.8",
        "Programming Language :: Python :: 3.9",
        "Topic :: Scientific/Engineering :: Physics",
        "Intended Audience :: Science/Research",
        "Development Status :: 5 - Production/Stable",
        "License :: OSI Approved :: GNU General Public License v2 or later "
        "(GPLv2+)"
        ],
    long_description=long_description,
    long_description_content_type="text/markdown",
    author_email="eugen.wintersberger@desy.de, dominik.kriegner@gmail.com",
    maintainer="Dominik Kriegner",
    maintainer_email="dominik.kriegner@gmail.com",
    package_dir={'': 'lib'},
    packages=find_packages('lib'),
    package_data={
        "xrayutilities": ["VERSION", "*.conf"],
        "xrayutilities.materials": [os.path.join("data", "*")]
        },
    python_requires='~=3.6',
    setup_requires=['numpy', 'scipy', 'h5py'],
    install_requires=['numpy>=1.9.2', 'scipy>=0.13.0', 'h5py'],
    extras_require={
        'plot': ["matplotlib>=3.1.0"],
        'fit': ["lmfit>=1.0.1"],
        },
    include_dirs=[numpy.get_include()],
    ext_modules=[extmodul],
    cmdclass=cmdclass,
    url="https://xrayutilities.sourceforge.io",
    license="GPLv2",
    script_args=args
    )



# This file is part of xrayutilities.
#
# xrayutilities is free software; you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation; either version 2 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program; if not, see <http://www.gnu.org/licenses/>.
#
# Copyright (C) 2015 Dominik Kriegner <dominik.kriegner@gmail.com>

import unittest

import numpy
import xrayutilities as xu


class TestFuzzyGridder2D(unittest.TestCase):
    @classmethod
    def setUpClass(cls):
        cls.n = 10
        cls.min = 1
        cls.max = 10
        cls.axis = numpy.linspace(cls.min, cls.max, cls.n)
        cls.x = int(numpy.random.rand()*10) + 1
        cls.y = int(numpy.random.rand()*10) + 1
        cls.data = numpy.random.rand()
        cls.gridder = xu.FuzzyGridder2D(cls.n, cls.n)
        cls.gridder.dataRange(cls.min, cls.max, cls.min, cls.max)
        cls.gridder(cls.x, cls.y, cls.data, width=2)

    def test_gridder2d_xaxis(self):
        # test length of xaxis
        self.assertEqual(len(self.gridder.xaxis), self.n)
        # test values of xaxis
        for i in range(self.n):
            self.assertAlmostEqual(self.gridder.xaxis[i], self.axis[i],
                                   places=12)

    def test_gridder2d_yaxis(self):
        # test length of yaxis
        self.assertEqual(len(self.gridder.yaxis), self.n)
        # test end values of yaxis
        for i in range(self.n):
            self.assertAlmostEqual(self.gridder.yaxis[i], self.axis[i],
                                   places=12)

    def test_fuzzygridder2d_data(self):
        # test shape of data
        self.assertEqual(self.gridder.data.shape[0], self.n)
        self.assertEqual(self.gridder.data.shape[1], self.n)
        # test values of data
        vg = numpy.zeros((self.gridder.data.shape))
        norm = numpy.copy(vg)
        ix, iy = self.x - 1, self.y - 1
        for i in range(ix-1, ix+2):
            for j in range(iy-1, iy+2):
                idx1 = i
                idx2 = j
                n = 1/4.
                if abs(i - ix) > 0:
                    n /= 2.
                if abs(j - iy) > 0:
                    n /= 2.
                if i < 0:
                    idx1 = 0
                elif i >= self.n:
                    idx1 = -1
                if j < 0:
                    idx2 = 0
                elif j >= self.n:
                    idx2 = -1
                norm[idx1, idx2] += n
                vg[idx1, idx2] = self.data

        for i in range(self.n):
            for j in range(self.n):
                self.assertAlmostEqual(self.gridder.data[i, j], vg[i, j],
                                       places=12)
                self.assertAlmostEqual(self.gridder._gnorm[i, j], norm[i, j],
                                       places=12)


if __name__ == '__main__':
    unittest.main()



# This file is part of xrayutilities.
#
# xrayutilities is free software; you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation; either version 2 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program; if not, see <http://www.gnu.org/licenses/>.
#
# Copyright (C) 2016-2017 Dominik Kriegner <dominik.kriegner@gmail.com>

import numbers
from math import sqrt

import numpy
from scipy import interpolate

from .. import utilities
from .powder import PowderDiffraction
from .smaterials import PowderList


class PowderModel(object):
    """
    Class to help with powder calculations for multiple materials.  For basic
    calculations the Powder class together with the Fundamental parameters
    approach is used.
    """

    def __init__(self, *args, **kwargs):
        """
        constructor for a powder model. The arguments consist of a PowderList
        or individual Powder(s). Optional parameters are specified in the
        keyword arguments.

        Note:
            After the end-of-use it is advisable to call the `close()` method
            to cleanup the multiprocessing calculation!

        Parameters
        ----------
        args :      PowderList or Powders
            either one PowderList or several Powder objects can be given
        kwargs :    dict
            optional parameters for the simulation. supported are:
        fpclass :   FP_profile, optional
            derived class with possible convolver mixins.  (default:
            FP_profile)
        fpsettings : dict
            settings dictionaries for the convolvers. Default settings are
            loaded from the config file.
        I0 :        float, optional
            scaling factor for the simulation result


        In particular interesting in fpsettings might be:
        {'displacement': {'specimen_displacement': z-displacement of the sample
                                                   from the rotation center
                          'zero_error_deg': zero error of the 2theta angle}
         'absorption': {'sample_thickness': sample thickness (m),
                        'absorption_coefficient': sample's absorption (m^-1)}
         'axial': {'length_sample': the length of the sample in the axial
                                    direction (m)}
        }
        """
        if len(args) == 1 and isinstance(args[0], PowderList):
            self.materials = args[0]
        else:
            self.materials = PowderList('%s List' % self.__class__.__name__,
                                        *args)
        self.I0 = kwargs.pop('I0', 1.0)
        self.pdiff = []
        kwargs['enable_simulation'] = True
        for mat in self.materials:
            self.pdiff.append(PowderDiffraction(mat, **kwargs))

        # default background
        self._bckg_type = 'polynomial'
        self._bckg_pol = [0, ]

    def set_parameters(self, params):
        """
        set simulation parameters of all subobjects

        Parameters
        ----------
        params :    dict
            settings dictionaries for the convolvers.
        """
        # set parameters for each convolver
        for pd in self.pdiff:
            pd.update_settings(params)

    def set_lmfit_parameters(self, lmparams):
        """
        function to update the settings of this class during an least squares
        fit

        Parameters
        ----------
        lmparams :  lmfit.Parameters
            lmfit Parameters list of sample and instrument parameters
        """
        pv = lmparams.valuesdict()
        settings = dict()
        h = list(self.pdiff[0].data)[0]
        fp = self.pdiff[0].data[h]['conv'].convolvers
        for conv in fp:
            name = conv[5:]
            settings[name] = dict()

        self.I0 = pv.pop('primary_beam_intensity', 1)
        set_splbkg = False
        spliney = {}
        for p in pv:
            if p.startswith('phase_'):  # sample phase parameters
                midx = 0
                for i, name in enumerate(self.materials.namelist):
                    if p.find(name) > 0:
                        midx = i
                name = self.materials.namelist[midx]
                attrname = p[p.find(name) + len(name) + 1:]
                setattr(self.materials[midx], attrname, pv[p])
            elif p.startswith('background_coeff'):
                self._bckg_pol[int(p.split('_')[-1])] = pv[p]
            elif p.startswith('background_spl_coeff'):
                set_splbkg = True
                spliney[int(p.split('_')[-1])] = pv[p]
            else:  # instrument parameters
                for k in settings:
                    if p.startswith(k):
                        slist = p[len(k) + 1:].split('_')
                        if len(slist) > 2 and slist[-2] == 'item':
                            name = '_'.join(slist[:-2])
                            if slist[-1] == '0':
                                settings[k][name] = []
                            settings[k][name].append(pv[p])
                        else:
                            name = p[len(k) + 1:]
                            settings[k][name] = pv[p]
                        break
        if set_splbkg:
            self._bckg_spline = interpolate.InterpolatedUnivariateSpline(
                self._bckg_spline._data[0],
                [spliney[k] for k in sorted(spliney)], ext=0)
        self.set_parameters(settings)

    def create_fitparameters(self):
        """
        function to create a fit model with all instrument and sample
        parameters.

        Returns
        -------
        lmfit.Parameters
        """
        lmfit = utilities.import_lmfit('XU.PowderModel')

        params = lmfit.Parameters()
        # sample phase parameters
        for mat, name in zip(self.materials, self.materials.namelist):
            for k in mat.__dict__:
                attr = getattr(mat, k)
                if isinstance(attr, numbers.Number):
                    params.add('_'.join(('phase', name, k)), value=attr,
                               vary=False)

        # instrument parameters
        settings = self.pdiff[0].settings
        for pg in settings:
            for p in settings[pg]:
                val = settings[pg][p]
                if p == 'dominant_wavelength' and pg == 'global':
                    # wavelength must be fit using emission_emiss_wavelength
                    continue
                if isinstance(val, numbers.Number):
                    params.add('_'.join((pg, p)), value=val, vary=False)
                elif isinstance(val, (numpy.ndarray, tuple, list)):
                    for j, item in enumerate(val):
                        params.add('_'.join((pg, p, 'item_%d' % j)),
                                   value=item, vary=False)

        # other global parameters
        params.add('primary_beam_intensity', value=self.I0, vary=False)
        if self._bckg_type == 'polynomial':
            for i, coeff in enumerate(self._bckg_pol):
                params.add('background_coeff_%d' % i, value=coeff, vary=False)
        elif self._bckg_type == 'spline':
            for i, coeff in enumerate(self._bckg_spline._data[1]):
                params.add('background_spl_coeff_%d' % i, value=coeff,
                           vary=False)
        return params

    def set_background(self, btype, **kwargs):
        """
        define background as spline or polynomial function

        Parameters
        ----------
        btype :     {polynomial', 'spline'}
            background type; Depending on this
            value the expected keyword arguments differ.
        kwargs :    dict
            optional keyword arguments
        x :     array-like, optional
            x-values (twotheta) of the background points (if btype='spline')
        y :     array-like, optional
            intensity values of the background (if btype='spline')
        p :     array-like, optional
            polynomial coefficients from the highest degree to the constant
            term. len of p decides about the degree of the polynomial (if
            btype='polynomial')
        """
        if btype == 'spline':
            self._bckg_spline = interpolate.InterpolatedUnivariateSpline(
                kwargs.get('x'), kwargs.get('y'), ext=0)
        elif btype == 'polynomial':
            self._bckg_pol = list(kwargs.get('p'))
        else:
            raise ValueError("btype must be either 'spline' or 'polynomial'")
        self._bckg_type = btype

    def simulate(self, twotheta, **kwargs):
        """
        calculate the powder diffraction pattern of all materials and sum the
        results based on the relative volume of the materials.

        Parameters
        ----------
        twotheta :  array-like
            positions at which the powder pattern should be evaluated
        kwargs :    dict
            optional keyword arguments
        background : array-like
            an array of background values (same shape as twotheta) if no
            background is given then the background is calculated as previously
            set by the set_background function or is 0.


        further keyword arguments are passed to the Convolve function of of the
        PowderDiffraction objects

        Returns
        -------
        array-like
            summed powder diffraction intensity of all materials present in the
            model
        """
        inte = numpy.zeros_like(twotheta)
        background = kwargs.pop('background', None)
        if background is None:
            if self._bckg_type == 'spline':
                background = self._bckg_spline(twotheta)
            else:
                background = numpy.polyval(self._bckg_pol, twotheta)
        totalvol = sum(pd.mat.volume for pd in self.pdiff)
        for pd in self.pdiff:
            inte += pd.Calculate(twotheta, **kwargs) * pd.mat.volume / totalvol
        return self.I0 * inte + background

    def fit(self, params, twotheta, data, std=None, maxfev=200):
        """
        make least squares fit with parameters supplied by the user

        Parameters
        ----------
        params :    lmfit.Parameters
            object with all parameters set as intended by the user
        twotheta :  array-like
            angular values for the fit
        data :      array-like
            experimental intensities for the fit
        std :       array-like
            standard deviation of the experimental data. if 'None' the sqrt of
            the data will be used
        maxfev:     int
            maximal number of simulations during the least squares refinement

        Returns
        -------
        lmfit.MinimizerResult
        """
        lmfit = utilities.import_lmfit('XU.PowderModel')

        def residual(pars, tt, data, weight):
            """
            residual function for lmfit Minimizer routine

            Parameters
            ----------
            pars :      lmfit.Parameters
                fit Parameters
            tt :        array-like
                array of twotheta angles
            data :      array-like
                experimental data, same shape as tt
            eps :       array-like
                experimental error bars, shape as tt
            """
            # set parameters in this instance
            self.set_lmfit_parameters(pars)

            # run simulation
            model = self.simulate(tt)
            return (model - data) * weight

        if std is None:
            weight = numpy.reciprocal(numpy.sqrt(data))
        else:
            weight = numpy.reciprocal(std)
        weight[numpy.isinf(weight)] = 1
        self.minimizer = lmfit.Minimizer(residual, params,
                                         fcn_args=(twotheta, data, weight))
        fitres = self.minimizer.minimize(max_nfev=maxfev)
        self.set_lmfit_parameters(fitres.params)
        return fitres

    def plot(self, twotheta, showlines=True, label='simulation', color=None,
             formatspec='-', lcolors=[], ax=None, **kwargs):
        """
        plot the powder diffraction pattern and indicate line positions for all
        components in the model.

        Parameters
        ----------
        twotheta :  array-like
            positions at which the powder pattern should be evaluated
        showlines : bool, optional
            flag to decide if peak positions of the components will be shown on
            the top of the plot
        label :     str
            line label in the plot
        color :     matplotlib color or None
            the color used for the line plot of the simulation
        formatspec : str
            format specifier of the simulation curve
        lcolors :   list of matplotlib colors
            colors for the line indicators for the various components
        ax :        matplotlib.axes or None
            axes object to be used for plotting, if its given no axes
            decoration like labels are set.

        Further keyword arguments are passed to the simulate method.

        Returns
        -------
        matplotlib.axes object
        """
        plot, plt = utilities.import_matplotlib_pyplot('XU.simpack')
        if not plot:
            return

        if ax is None:
            fig, iax = plt.subplots()
            iax.set_xlabel(r"$2\theta$ (°)")
            iax.set_ylabel(r"Intensity")
            iax.set_xlim(twotheta.min(), twotheta.max())
        else:
            fig = ax.figure
            iax = ax

        plotkwargs = dict(label=label)
        if color is not None:
            plotkwargs['color'] = color
        iax.plot(twotheta, self.simulate(twotheta, **kwargs), formatspec,
                 **plotkwargs)

        if showlines:
            from matplotlib.colors import is_color_like
            from mpl_toolkits.axes_grid1 import make_axes_locatable
            divider = make_axes_locatable(iax)
            taxlist = []
            lineslist = []
            annotlist = []
            settings = self.pdiff[0].settings
            wavelengths = settings['emission']['emiss_wavelengths']
            intensities = settings['emission']['emiss_intensities']
            for i, pd in enumerate(self.pdiff):
                tax = divider.append_axes("top", size="6%", pad=0.05,
                                          sharex=iax)
                if lcolors:
                    c = lcolors[i % len(lcolors)]
                elif len(self.pdiff) == 1:
                    if is_color_like(color):
                        c = color
                    elif is_color_like(formatspec[-1]):
                        c = formatspec[-1]
                    else:
                        c = 'C0'
                else:
                    c = f'C{i}'
                lw = 2
                wllist = []
                for wl, inte in zip(wavelengths, intensities):
                    q = [pd.data[h]['qpos'] for h in pd.data]
                    tt = pd.Q2Ang(q, wl=1e10*wl)*2
                    lw *= inte
                    lines = tax.vlines(tt, 0, 1, colors=c, linewidth=lw)
                    wllist.append(lines)

                plt.setp(tax.get_xticklabels(), visible=False)
                plt.setp(tax.get_yticklabels(), visible=False)
                plt.setp(tax.get_yticklines(), visible=False)
                annot = tax.annotate(
                    "", xy=(0, 0), xytext=(20, 0), textcoords="offset points",
                    bbox=dict(boxstyle="round,pad=0.1", fc="w", alpha=0.8),
                    arrowprops=dict(arrowstyle="->"), fontsize='x-small')
                annot.set_visible(False)
                # next line important to avoid zorder issues
                tax.figure.texts.append(tax.texts.pop())
                taxlist.append(tax)
                lineslist.append(wllist)
                annotlist.append(annot)

            def update_annot(pd, annot, lines, ind):
                h = list(pd.data)[ind]
                x = 2*pd.data[h]['ang']
                y = 0.5
                annot.xy = (x, y)
                text = "{}: {} {} {}".format(pd.mat.name, h[0], h[1], h[2])
                annot.set_text(text)
                annot.get_bbox_patch().set_edgecolor(lines.get_color()[0])
                annot.set_zorder(10)

            def hover(event):
                for pd, tax, annot, wllist in zip(self.pdiff, taxlist,
                                                  annotlist, lineslist):
                    vis = annot.get_visible()
                    if event.inaxes == tax:
                        for lines in wllist:
                            cont, ind = lines.contains(event)
                            if cont:
                                update_annot(pd, annot, lines, ind['ind'][0])
                                annot.set_visible(True)
                                fig.canvas.draw_idle()
                                return
                            else:
                                if vis:
                                    annot.set_visible(False)
                                    fig.canvas.draw_idle()
                                    return

            def click(event):
                for pd, tax, annot, wllist in zip(self.pdiff, taxlist,
                                                  annotlist, lineslist):
                    if event.inaxes == tax:
                        for lines in wllist:
                            cont, ind = lines.contains(event)
                            if cont:
                                h = list(pd.data)[ind['ind'][0]]
                                text = (f'{pd.mat.name}: {h[0]} {h[1]} {h[2]};'
                                        f' 2Theta = ')
                                for wl in wavelengths:
                                    tt = 2 * pd.Q2Ang(pd.data[h]['qpos'],
                                                      wl=1e10*wl)
                                    text += f'{tt:.4f}°, '
                                print(text[:-2])
                                return

            fig.canvas.mpl_connect("motion_notify_event", hover)
            fig.canvas.mpl_connect("button_press_event", click)
        if ax is None:
            fig.tight_layout()
        return iax

    def close(self):
        for pd in self.pdiff:
            pd.close()

    def __str__(self):
        """
        string representation of the PowderModel
        """
        ostr = "PowderModel {\n"
        ostr += "I0: %f\n" % self.I0
        ostr += str(self.materials)
        ostr += "}"
        return ostr


def Rietveld_error_metrics(exp, sim, weight=None, std=None,
                           Nvar=0, disp=False):
    """
    calculates common error metrics for Rietveld refinement.

    Parameters
    ----------
    exp :       array-like
        experimental datapoints
    sim :       array-like
        simulated data
    weight :    array-like, optional
        weight factor in the least squares sum. If it is None the weight is
        estimated from the counting statistics of 'exp'
    std :       array-like, optional
        standard deviation of the experimental data. alternative way of
        specifying the weight factor. when both are given weight overwrites
        std!
    Nvar :      int, optional
        number of variables in the refinement
    disp :      bool, optional
        flag to tell if a line with the calculated values should be printed.

    Returns
    -------
    M, Rp, Rwp, Rwpexp, chi2: float
    """
    if weight is None and std is None:
        weight = numpy.reciprocal(exp)
    elif weight is None:
        weight = numpy.reciprocal(std**2)
    weight[numpy.isinf(weight)] = 1
    M = numpy.sum((exp - sim)**2 * weight)
    Rp = numpy.sum(numpy.abs(exp - sim))/numpy.sum(exp)
    Rwp = sqrt(M / numpy.sum(weight * exp**2))
    chi2 = M / (len(exp) - Nvar)
    Rwpexp = Rwp / sqrt(chi2)
    if disp:
        print('Rp=%.4f Rwp=%.4f Rwpexp=%.4f chi2=%.4f'
              % (Rp, Rwp, Rwpexp, chi2))
    return M, Rp, Rwp, Rwpexp, chi2


def plot_powder(twotheta, exp, sim, mask=None, scale='sqrt', fig='XU:powder',
                show_diff=True, show_legend=True, labelexp='experiment',
                labelsim='simulation', formatexp='.-k', formatsim='-r'):
    """
    Convenience function to plot the comparison between experimental and
    simulated powder diffraction data

    Parameters
    ----------
    twotheta :  array-like
        angle values used for the x-axis of the plot (deg)
    exp :       array-like
        experimental data (same shape as twotheta). If None only the simulation
        and no difference will be plotted
    sim :       array-like or PowederModel
        simulated data or PowderModel instance. If a PowderModel instance is
        given the plot-method of PowderModel is used.
    mask :      array-like, optional
        mask to reduce the twotheta values to the be used as x-coordinates of
        sim
    scale :     {'linear', 'sqrt', 'log'}, optional
        string specifying the scale of the y-axis.
    fig :       str or int, optional
        matplotlib figure name (figure will be cleared!)
    show_diff : bool, optional
        flag to specify if a difference curve should be shown
    show_legend: bool, optional
        flag to specify if a legend should be shown
    labelexp :  str
        plot label (legend entry) for the experimental data
    labelsim :  str
        plot label for the simulation data
    formatexp : str
        format specifier for the experimental data
    formatsim : str
        format specifier for the simulation curve
    """
    plot, plt = utilities.import_matplotlib_pyplot('XU.simpack')
    if not plot:
        return

    f = plt.figure(fig, figsize=(10, 7))
    f.clf()
    ax = plt.subplot(111)
    if exp is not None:
        ax.plot(twotheta, exp, formatexp, label=labelexp)
    if mask is None:
        mask = numpy.ones_like(twotheta, dtype=numpy.bool)
    if isinstance(sim, PowderModel):
        simdata = sim.simulate(twotheta[mask])
        sim.plot(twotheta[mask], label=labelsim, formatspec=formatsim, ax=ax)
    else:
        simdata = sim
        ax.plot(twotheta[mask], simdata, formatsim, label=labelsim)

    if show_diff and exp is not None:
        # plot error between simulation and experiment
        ax.plot(twotheta[mask], exp[mask]-simdata, '.-', color='0.5',
                label='difference')

    ax.set_xlabel('2Theta (deg)')
    ax.set_ylabel('Intensity')
    lines = ax.get_lines()
    plt.figlegend(lines, [line.get_label() for line in lines],
                  loc='upper right', frameon=True)
    ax.set_yscale(scale)
    f.tight_layout()
    return lines



# This file is part of xrayutilities.
#
# xrayutilities is free software; you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation; either version 2 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program; if not, see <http://www.gnu.org/licenses/>.
#
# Copyright (C) 2015-2020 Dominik Kriegner <dominik.kriegner@gmail.com>

import math
import os
import tempfile
import unittest

import xrayutilities as xu


class TestDatabase(unittest.TestCase):
    @classmethod
    def setUpClass(cls):
        cls.el = xu.materials.elements.Dummy
        # test creation of database
        fid, cls.dbfilename = tempfile.mkstemp(suffix='.db')
        os.close(fid)
        xu.materials.database.createAndFillDatabase(cls.dbfilename)
        cls.db = xu.materials.DataBase(cls.dbfilename)
        cls.db.Open()
        cls.db.SetMaterial(cls.el.name)

    @classmethod
    def tearDownClass(cls):
        cls.db.Close()
        try:
            os.remove(cls.dbfilename)
        except OSError:
            print("OSError during cleaning up of temporary database file!")
            pass

    def test_db_f0(self):
        f0 = self.el.f0(0)
        self.assertAlmostEqual(f0, 1.0, places=10)

    def test_owndb_f0(self):
        f0 = self.db.GetF0(0)
        self.assertAlmostEqual(f0, 1.0, places=10)

    def test_db_f1_neg(self):
        f1 = self.el.f1(-1)
        self.assertTrue(math.isnan(f1))

    def test_owndb_f1_neg(self):
        f1 = self.db.GetF1(-1)
        self.assertTrue(math.isnan(f1))

    def test_db_f1(self):
        f1 = self.el.f1(1000)
        self.assertAlmostEqual(f1, 0.0, places=10)

    def test_owndb_f1(self):
        f1 = self.db.GetF1(1000)
        self.assertAlmostEqual(f1, 0.0, places=10)

    def test_db_f2_neg(self):
        f2 = self.el.f2(-1)
        self.assertTrue(math.isnan(f2))

    def test_owndb_f2_neg(self):
        f2 = self.db.GetF2(-1)
        self.assertTrue(math.isnan(f2))

    def test_db_f2(self):
        f2 = self.el.f2(1000)
        self.assertAlmostEqual(f2, 0.0, places=10)

    def test_owndb_f2(self):
        f2 = self.db.GetF2(1000)
        self.assertAlmostEqual(f2, 0.0, places=10)


if __name__ == '__main__':
    unittest.main()



# This file is part of xrayutilities.
#
# xrayutilities is free software; you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation; either version 2 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program; if not, see <http://www.gnu.org/licenses/>.
#
# Copyright (C) 2015 Dominik Kriegner <dominik.kriegner@gmail.com>

import os.path
import unittest

import numpy
import xrayutilities as xu

xu.config.VERBOSITY = 0  # make no output during test
testfile = 'rigaku_rsm.ras.gz'
datadir = os.path.join(os.path.dirname(__file__), 'data')
fullfilename = os.path.join(datadir, testfile)


@unittest.skipIf(not os.path.isfile(fullfilename),
                 "additional test data needed (http://xrayutilities.sf.io)")
class TestIO_Rigaku(unittest.TestCase):
    nscans = 401
    dshape = (301,)
    scannr = 21
    dmax = 187.0
    dmin = 101.0
    motmax = 29.90
    motmin = 26.90
    tpos = 34
    dtpos = 164.0
    motorname = 'TwoTheta'
    countername = 'int'

    @classmethod
    def setUpClass(cls):
        cls.rasfile = xu.io.RASFile(testfile, path=datadir)
        cls.sdata = cls.rasfile.scans[cls.scannr].data
        cls.motor, data = xu.io.getras_scan(testfile+'%s', '',
                                            cls.motorname, path=datadir)
        cls.inte = data[cls.countername]

    def test_datashape_ras(self):
        self.assertEqual(self.nscans, len(self.rasfile.scans))
        self.assertEqual(self.dshape, self.sdata.shape)

    def test_datashape_all(self):
        self.assertEqual(self.dshape[0]*self.nscans, self.inte.size)
        self.assertEqual(self.dshape[0]*self.nscans, self.motor.size)

    def test_datavalues(self):
        mot = self.sdata[self.motorname]
        inte = self.sdata[self.countername]
        self.assertAlmostEqual(self.motmax, mot.max(), places=6)
        self.assertAlmostEqual(self.motmin, mot.min(), places=6)
        self.assertAlmostEqual(self.dmax, inte.max(), places=6)
        self.assertAlmostEqual(self.dmin, inte.min(), places=6)
        self.assertAlmostEqual(self.dtpos, inte[self.tpos], places=6)

    def test_datamethods(self):
        self.assertTrue(numpy.all(self.sdata[self.countername] ==
                                  self.inte[self.scannr*self.dshape[0]:
                                            (self.scannr+1)*self.dshape[0]]))


if __name__ == '__main__':
    unittest.main()



# This file is part of xrayutilities.
#
# xrayutilities is free software; you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation; either version 2 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program; if not, see <http://www.gnu.org/licenses/>.
#
# Copyright (C) 2012-2020 Dominik Kriegner <dominik.kriegner@gmail.com>

import matplotlib.pyplot as plt
import numpy
import xrayutilities as xu

# defining material and experimental setup
InAs = xu.materials.InAs
energy = numpy.linspace(500, 20000, 5000)  # 500 - 20000 eV

F = InAs.StructureFactorForEnergy(InAs.Q(1, 1, 1), energy)

plt.figure()
plt.clf()
plt.plot(energy, F.real, '-k', label='Re(F)')
plt.plot(energy, F.imag, '-r', label='Imag(F)')
plt.xlabel("Energy (eV)")
plt.ylabel("F")
plt.legend()



# This file is part of xrayutilities.
#
# xrayutilities is free software; you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation; either version 2 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program; if not, see <http://www.gnu.org/licenses/>.
#
# Copyright (C) 2014 Dominik Kriegner <dominik.kriegner@gmail.com>

import unittest

import numpy
import xrayutilities as xu


class TestBlockAverageFunctions(unittest.TestCase):

    def setUp(self):
        self.seq = numpy.random.rand(11)
        self.n = 3
        self.seq2d = numpy.random.rand(10, 15)
        self.n2d = (3, 4)

    def test_blockav1d(self):
        out = xu.blockAverage1D(self.seq, self.n)
        self.assertAlmostEqual(out[0], numpy.average(self.seq[0:self.n]))
        self.assertEqual(out.size, numpy.ceil(self.seq.size / float(self.n)))

    def test_blockav2d(self):
        out = xu.blockAverage2D(self.seq2d, self.n2d[0], self.n2d[1])
        self.assertAlmostEqual(
            out[0, 0],
            numpy.average(self.seq2d[0:self.n2d[0], 0:self.n2d[1]]))
        self.assertEqual(
            out.shape,
            (numpy.ceil(self.seq2d.shape[0] / float(self.n2d[0])),
             numpy.ceil(self.seq2d.shape[1] / float(self.n2d[1]))))

    def test_blockav_psd(self):
        out = xu.blockAveragePSD(self.seq2d, self.n)
        self.assertAlmostEqual(out[0, 0],
                               numpy.average(self.seq2d[0, 0:self.n]))
        self.assertEqual(
            out.shape,
            (self.seq2d.shape[0],
             numpy.ceil(self.seq2d.shape[1] / float(self.n))))


if __name__ == '__main__':
    unittest.main()



# This file is part of xrayutilities.
#
# xrayutilities is free software; you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation; either version 2 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program; if not, see <http://www.gnu.org/licenses/>.
#
# Copyright (C) 2016 Dominik Kriegner <dominik.kriegner@gmail.com>

import unittest

import xrayutilities as xu


class Test_maplog(unittest.TestCase):
    @classmethod
    def setUpClass(cls):
        cls.a = xu.materials.Amorphous('Ir0.2Mn0.8', 10130)

    def test_elements(self):
        self.assertEqual(self.a.base[0][0], xu.materials.elements.Ir)
        self.assertEqual(self.a.base[1][0], xu.materials.elements.Mn)

    def test_composition(self):
        self.assertAlmostEqual(self.a.base[0][1], 0.2, places=10)
        self.assertAlmostEqual(self.a.base[1][1], 0.8, places=10)


if __name__ == '__main__':
    unittest.main()



# This file is part of xrayutilities.
#
# xrayutilities is free software; you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation; either version 2 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program; if not, see <http://www.gnu.org/licenses/>.
#
# Copyright (C) 2015 Dominik Kriegner <dominik.kriegner@gmail.com>

import os.path
import unittest

import xrayutilities as xu

testfile = 'pilatus100K.tif'
datadir = os.path.join(os.path.dirname(__file__), 'data')
fullfilename = os.path.join(datadir, testfile)


@unittest.skipIf(not os.path.isfile(fullfilename),
                 "additional test data needed (http://xrayutilities.sf.io)")
class TestIO_Pilatus(unittest.TestCase):
    dshape = (195, 487)
    dmax = 1.0
    dmin = -2.0
    tpos = (75, 281)
    dtpos = 1.0

    @classmethod
    def setUpClass(cls):
        cls.data = xu.io.get_tiff(testfile, path=datadir)

    def test_datashape(self):
        self.assertEqual(self.dshape, self.data.shape)

    def test_datavalues(self):
        self.assertAlmostEqual(self.dmax, self.data.max(), places=10)
        self.assertAlmostEqual(self.dmin, self.data.min(), places=10)
        self.assertAlmostEqual(self.dtpos,
                               self.data[self.tpos[0], self.tpos[1]],
                               places=10)


if __name__ == '__main__':
    unittest.main()



# This file is part of xrayutilities.
#
# xrayutilities is free software; you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation; either version 2 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program; if not, see <http://www.gnu.org/licenses/>.
#
# Copyright (C) 2009-2010 Eugen Wintersberger <eugen.wintersberger@desy.de>
# Copyright (C) 2009-2019 Dominik Kriegner <dominik.kriegner@gmail.com>

from .cbf import CBFDirectory, CBFFile
from .desy_tty08 import gettty08_scan, tty08File
from .edf import EDFDirectory, EDFFile
from .fastscan import FastScan, FastScanCCD, FastScanSeries
from .helper import xu_h5open, xu_open
from .ill_numor import numor_scan, numorFile
from .imagereader import (ImageReader, PerkinElmer, Pilatus100K, RoperCCD,
                          TIFFRead, get_tiff)
from .panalytical_xml import XRDMLFile, getxrdml_map, getxrdml_scan
from .pdcif import pdCIF, pdESG
from .rigaku_ras import RASFile, RASScan, getras_scan
# parser for the alignment log file of the rotating anode
from .rotanode_alignment import RA_Alignment
from .seifert import SeifertMultiScan, SeifertScan, getSeifert_map
from .spec import SPECFile, SPECLog, SPECScan, geth5_scan, getspec_scan
from .spectra import SPECTRAFile, geth5_spectra_map



# This file is part of xrayutilities.
#
# xrayutilities is free software; you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation; either version 2 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program; if not, see <http://www.gnu.org/licenses/>.
#
# Copyright (C) 2014 Dominik Kriegner <dominik.kriegner@gmail.com>

import unittest

import numpy
import xrayutilities as xu


class TestQConversionTrans(unittest.TestCase):
    @classmethod
    def setUpClass(cls):
        cls.nch = 9
        cls.ncch = 4
        cls.nch2d = (9, 13)
        cls.ncch1 = 4
        cls.ncch2 = 6
        # standard 1S+1D goniometer
        qconv = xu.QConversion('x+', 'x+', (0, 1, 0))
        cls.hxrd = xu.HXRD((1., 1., 0.), (0., 0., 1.), qconv=qconv)
        # comparable goniometer with translations
        qconv = xu.QConversion('x+', ['ty', 'tz'], (0, 1e-15, 0))
        cls.hxrdtrans = xu.HXRD((1., 1., 0.), (0., 0., 1.), qconv=qconv,
                                sampleor='z+')
        cls.hxrdtrans.Ang2Q.init_linear('z+', cls.ncch, cls.nch, 1e-15, 50e-6)
        cls.hxrdtrans.Ang2Q.init_area('z+', 'x+', cls.ncch1, cls.ncch2,
                                      cls.nch2d[0], cls.nch2d[1],
                                      1e-15, 50e-6, 50e-6)

        cls.angle = numpy.random.rand() * 45

    def test_qtrans0(self):
        qvec1 = self.hxrd.Ang2Q(self.angle, 0)
        qvec2 = self.hxrdtrans.Ang2Q(self.angle, 1, 0)
        for i in range(3):
            self.assertAlmostEqual(qvec1[i], qvec2[i], places=10)

    def test_qtrans45(self):
        qvec1 = self.hxrd.Ang2Q(self.angle, 45)
        qvec2 = self.hxrdtrans.Ang2Q(self.angle, 1, 1)

        for i in range(3):
            self.assertAlmostEqual(qvec1[i], qvec2[i], places=10)

    def test_qtrans90(self):
        qvec1 = self.hxrd.Ang2Q(self.angle, 90)
        qvec2 = self.hxrdtrans.Ang2Q(self.angle, 0, 1)
        for i in range(3):
            self.assertAlmostEqual(qvec1[i], qvec2[i], places=10)

    def test_qtrans0_linear(self):
        qvec1 = self.hxrd.Ang2Q(self.angle, 0)
        qx, qy, qz = self.hxrdtrans.Ang2Q.linear(self.angle, 1, 0)
        qvec2 = (qx[self.ncch], qy[self.ncch], qz[self.ncch])
        for i in range(3):
            self.assertAlmostEqual(qvec1[i], qvec2[i], places=10)

    def test_qtrans45_linear(self):
        qvec1 = self.hxrd.Ang2Q(self.angle, 45)
        qx, qy, qz = self.hxrdtrans.Ang2Q.linear(self.angle, 1, 1)
        qvec2 = (qx[self.ncch], qy[self.ncch], qz[self.ncch])
        for i in range(3):
            self.assertAlmostEqual(qvec1[i], qvec2[i], places=10)

    def test_qtrans90_linear(self):
        qvec1 = self.hxrd.Ang2Q(self.angle, 90)
        qx, qy, qz = self.hxrdtrans.Ang2Q.linear(self.angle, 0, 1)
        qvec2 = (qx[self.ncch], qy[self.ncch], qz[self.ncch])
        for i in range(3):
            self.assertAlmostEqual(qvec1[i], qvec2[i], places=10)

    def test_qtrans0_area(self):
        qvec1 = self.hxrd.Ang2Q(self.angle, 0)
        qx, qy, qz = self.hxrdtrans.Ang2Q.area(self.angle, 1, 0)
        qvec2 = (qx[self.ncch1, self.ncch2],
                 qy[self.ncch1, self.ncch2],
                 qz[self.ncch1, self.ncch2])
        for i in range(3):
            self.assertAlmostEqual(qvec1[i], qvec2[i], places=10)

    def test_qtrans45_area(self):
        qvec1 = self.hxrd.Ang2Q(self.angle, 45)
        qx, qy, qz = self.hxrdtrans.Ang2Q.area(self.angle, 1, 1)
        qvec2 = (qx[self.ncch1, self.ncch2],
                 qy[self.ncch1, self.ncch2],
                 qz[self.ncch1, self.ncch2])
        for i in range(3):
            self.assertAlmostEqual(qvec1[i], qvec2[i], places=10)

    def test_qtrans90_area(self):
        qvec1 = self.hxrd.Ang2Q(self.angle, 90)
        qx, qy, qz = self.hxrdtrans.Ang2Q.area(self.angle, 0, 1)
        qvec2 = (qx[self.ncch1, self.ncch2],
                 qy[self.ncch1, self.ncch2],
                 qz[self.ncch1, self.ncch2])
        for i in range(3):
            self.assertAlmostEqual(qvec1[i], qvec2[i], places=10)


if __name__ == '__main__':
    unittest.main()



# This file is part of xrayutilities.
#
# xrayutilities is free software; you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation; either version 2 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program; if not, see <http://www.gnu.org/licenses/>.
#
# Copyright (C) 2017-2020 Dominik Kriegner <dominik.kriegner@gmail.com>

import math
import unittest

import numpy
import xrayutilities as xu
from numpy import arccos, cos, radians, sin, sqrt


class TestMaterialsTransform(unittest.TestCase):
    @classmethod
    def setUp(cls):
        cls.a, cls.b, cls.c = numpy.random.rand(3) * 2 + 4
        cls.alpha, cls.beta, cls.gamma = numpy.random.rand(3) * 60 + 60
        cls.p1mat = xu.materials.Crystal(
            'P1', xu.materials.SGLattice(1, cls.a, cls.b, cls.c,
                                         cls.alpha, cls.beta, cls.gamma))

    def test_q2hkl_hkl2q(self):
        for i in range(3):
            hkls = numpy.random.randint(-5, 6, 3)
            qvec = self.p1mat.Q(hkls)
            backhkl = self.p1mat.HKL(qvec)
            for j in range(3):
                self.assertAlmostEqual(hkls[j], backhkl[j], places=10)

    def test_Bmatrix(self):
        """
        check if our B matrix is compatible with the one from
        Busing&Levy Acta Cryst. 22, 457 (1967)
        """
        ca = cos(radians(self.alpha))
        cb = cos(radians(self.beta))
        cg = cos(radians(self.gamma))
        sa = sin(radians(self.alpha))
        sb = sin(radians(self.beta))
        sg = sin(radians(self.gamma))
        vh = sqrt(1 - ca**2-cb**2-cg**2 + 2*ca*cb*cg)
        pi2 = numpy.pi * 2
        ra, rb, rc = pi2*sa/(self.a*vh), pi2*sb/(self.b*vh), pi2*sg/(self.c*vh)
        # cralpha = (cb*cg - ca)/(sb*sg)
        crbeta = (ca*cg - cb)/(sa*sg)
        crgamma = (ca*cb - cg)/(sa*sb)

        b = numpy.zeros((3, 3))
        b[0, 0] = ra
        b[0, 1] = rb * crgamma
        b[1, 1] = rb * sin(arccos(crgamma))
        b[0, 2] = rc * crbeta
        b[1, 2] = -rc * sin(arccos(crbeta)) * ca
        b[2, 2] = pi2 / self.c

        for j in range(9):
            self.assertAlmostEqual(b.flat[j], self.p1mat.B.flat[j], places=10)

    def test_Bmatrix_after_a_setter(self):
        self.a = numpy.random.rand() * 2 + 4
        self.p1mat.lattice.a = self.a
        self.test_Bmatrix()

    def test_Bmatrix_after_alpha_setter(self):
        self.alpha = numpy.random.rand() * 60 + 60
        self.p1mat.lattice.alpha = self.alpha
        self.test_Bmatrix()

    def test_Bmatrix_after_all_setters(self):
        # change materials unit cell parameters to test setters
        self.a, self.b, self.c = numpy.random.rand(3) * 2 + 4
        self.alpha, self.beta, self.gamma = numpy.random.rand(3) * 60 + 60
        self.p1mat.lattice.a = self.a
        self.p1mat.lattice.b = self.b
        self.p1mat.lattice.c = self.c
        self.p1mat.lattice.alpha = self.alpha
        self.p1mat.lattice.beta = self.beta
        self.p1mat.lattice.gamma = self.gamma
        self.test_Bmatrix()

    def test_environment(self):
        maxdist = max(self.a, self.b, self.c) + 0.01
        e = self.p1mat.environment((0, 0, 0), maxdist=maxdist)

        self.assertTrue(len(e) >= 4,
                        f"Length of environment must be >= 4, is {len(e)}")
        for dis in (0.0, self.a, self.b, self.c):
            found = False
            for d, at, mult in e:
                if numpy.isclose(d, dis):
                    found = True
            self.assertTrue(found, "expected atomic distance not found")

    def test_environment_Si(self):
        a = xu.materials.Si.a
        e = xu.materials.Si.environment(0.125, 0.125, 0.125)

        self.assertAlmostEqual(e[0][0], a*math.sqrt(3)/8, places=10)
        self.assertEqual(e[0][1], xu.materials.elements.Si)
        self.assertAlmostEqual(e[0][2], 2.0)

    def test_isequivalent(self):
        hkl1 = (1, 2, 3)
        materials = ['C', 'C_HOPG', 'TiO2', 'GeTe', 'Ag2Se']
        hkl2lst = [((2, 1, -3), (2, 2, 3)),
                   ((1, -3, 3), (1, -2, 3)),
                   ((-2, 1, 3), (3, 2, 1)),
                   ((1, 3, 2), (1, 3, -2)),
                   ((1, -2, -3), (1, 3, 2))]
        for mname, hkl2s in zip(materials, hkl2lst):
            mat = getattr(xu.materials, mname)
            self.assertTrue(mat.lattice.isequivalent(hkl1, hkl2s[0]))
            self.assertFalse(mat.lattice.isequivalent(hkl1, hkl2s[1]))


if __name__ == '__main__':
    unittest.main()
